MODULE first_guess_map

   USE namelist_info
   USE constants

   REAL , PARAMETER      :: radians_per_degree = pi / 180.

CONTAINS

   SUBROUTINE find_center_longitude_lambert ( grid_distance , cone_factor , known_longitude , &
                                              center_longitude , true_latitude , known_latitude , &
                                              known_x , point_x )  
      
      IMPLICIT NONE
   
      !  Input data.
   
      REAL , INTENT(IN)     :: known_latitude  , &  ! this value of latitude we know, usually lower left point
                               known_longitude , &  ! this value of longitude we know, usually lower left point
                               cone_factor     , &  ! the cone factor of the projection
                               true_latitude   , &  ! the first true latitude
                               center_longitude, &  ! the longitude that is parallel to the y-axis
                               known_x         , &  ! x location of the known value of latitude, usually 1
                               grid_distance        ! grid distance at true latitude
   
      !  Output data.
   
      REAL , INTENT(OUT)    :: point_x              ! x location of point with unknown latitude, usually center
   
      !  Using the input information, compute the X location for the center of
      !  the projection.  
   
      !  Local data.
   
      REAL                  :: xx , r
      REAL                  :: known_longitude_hold

      !  If the center longitude and the known longitude are in different hemispheres, fix that so that the
      !  distance is < 180 degrees.

      known_longitude_hold = known_longitude
      
      if(known_longitude * center_longitude .LT. 0 ) THEN
         IF ( center_longitude .LT. 0 ) THEN
            known_longitude_hold = known_longitude - 360
         ELSE
            known_longitude_hold = known_longitude + 360
         END IF
      END IF
   
      !  This is unintelligible, but if it is right, who is going to try to 
      !  understand it anyways?
   
      r=radius_of_earth/cone_factor * &
          SIN((90.-true_latitude)*radians_per_degree) * &
        ( TAN((90.-abs(known_latitude))*radians_per_degree/2.) / &
          TAN((90.-abs( true_latitude))*radians_per_degree/2.) )**cone_factor

      xx=r*SIN(cone_factor*(known_longitude_hold-center_longitude)*radians_per_degree)
   
      point_x=known_x - xx/grid_distance
   
   END SUBROUTINE find_center_longitude_lambert
   
   SUBROUTINE find_center_longitude_polar ( grid_distance , known_longitude , &
                                            center_longitude , true_latitude , known_latitude , &
                                            known_x , point_x )  
      
      IMPLICIT NONE
   
      !  Input data.
   
      REAL , INTENT(IN)     :: known_latitude  , &  ! this value of latitude we know, usually lower left point
                               known_longitude , &  ! this value of longitude we know, usually lower left point
                               true_latitude   , &  ! the first true latitude
                               center_longitude, &  ! the longitude that is parallel to the y-axis
                               known_x         , &  ! x location of the known value of latitude, usually 1
                               grid_distance        ! grid distance at true latitude
   
      !  Output data.
   
      REAL , INTENT(OUT)    :: point_x              ! x location of point with unknown latitude, usually center
   
      !  Using the input information, compute the X location for the center of
      !  the projection.  
   
      !  Local data.
   
      REAL                  :: f , r
   
      !  This is unintelligible, but if it is right, who is going to try to 
      !  understand it anyways?  The values for true latitude should probably
      !  be set to permit data from south of the equator.
   
      r=radius_of_earth *SIN((90.-known_latitude)*radians_per_degree) * &
                     (1.+COS((90.- true_latitude)*radians_per_degree)) / &
                     (1.+COS((90.-known_latitude)*radians_per_degree))
   
      f = ( known_longitude - center_longitude - 90. ) * radians_per_degree

      point_x=known_x - r*COS(f)/grid_distance
   
   END SUBROUTINE find_center_longitude_polar

   SUBROUTINE center_latitude_lambert ( known_latitude , cone_factor , true_latitude , center_longitude , &
                                        known_x , point_x , known_y , point_y , &
                                        grid_distance , point_latitude , converged )
   
      IMPLICIT NONE
   
      !  Input data.
   
      REAL , INTENT(IN)     :: known_latitude  , &  ! this value of latitude we know, usually lower left point
                               cone_factor     , &  ! the cone factor of the projection
                               true_latitude   , &  ! the first true latitude
                               center_longitude, &  ! the longitude that is parallel to the y-axis
                               known_x         , &  ! x location of the known value of latitude, usually 1
                               point_x         , &  ! x location of point with unknown latitude, usually center
                               known_y         , &  ! y location of the known value of latitude, usually 1 
                               point_y         , &  ! y location of point with unknown latitude, usually center
                               grid_distance        ! grid distance at true latitude
   
      !  Output data.
   
      REAL , INTENT(OUT)    :: point_latitude       ! the value of the latitude at point_x, point_y
      LOGICAL , INTENT(OUT) :: converged
   
      !  Local data.
   
      REAL                  :: c1 , xlat , r , &
                               rr , xa , xb, xn , fa , fb , fn , xo
   
      INTEGER               :: loop , max_loops = 100
   
      LOGICAL               :: error = .FALSE.
   
   
      !  This routine uses the bisection method to determine the latitude of 
      !  the center of the domain.  The input data is typical of what is available
      !  from the GRIB files for a Lambert conformal projection.  
   
      !  Statement functions for use in the iteration.
   
      xlat(rr,cone_factor,true_latitude) = SIGN(90.,true_latitude) - &
      ( 2. * ATAN(((rr * cone_factor)/ &
                   (SIN((ABS(SIGN(90.,true_latitude)-true_latitude))*radians_per_degree)*radius_of_earth))**(1./cone_factor) * &
      TAN((SIGN(90.,true_latitude)-true_latitude)/2.*radians_per_degree) ) ) / radians_per_degree
   
      r(known_x,point_x,known_y,point_y,grid_distance,c1,point_latitude)= SQRT ( &
      ((known_x-point_x)*grid_distance)**2 + &
      ((known_y-point_y)*grid_distance - SIGN(1.,true_latitude)*&
      (c1*TAN(((ABS(SIGN(90.,true_latitude)-point_latitude))*radians_per_degree)/2.)**cone_factor))**2 ) 
   
      !  Local constants from input data and PARAMETERs.  Initialize the output
      !  value to a recognizably undefined latitude.
   
      c1 = radius_of_earth * SIN((ABS(SIGN(90.,true_latitude)-true_latitude))    * radians_per_degree) / &
           ( cone_factor * ( TAN((ABS(SIGN(90.,true_latitude)-true_latitude))/2. * radians_per_degree) ) ** cone_factor ) 
      point_latitude = -999. 
   
      !  Compute the starting values of the bisection.
       
      xa = SIGN( 1.,true_latitude)
      fa = xlat(r(known_x,point_x,known_y,point_y,grid_distance,c1,xa), &
           cone_factor,true_latitude)

      xb = SIGN(89.,true_latitude)
      fb = xlat(r(known_x,point_x,known_y,point_y,grid_distance,c1,xb), &
           cone_factor,true_latitude)
   
      !  First test to see that we have chosen reasonable bounds.
   
      IF ( ( fa - known_latitude ) * ( fb - known_latitude ) .LE. 0 ) THEN
         IF ( rec4%print_debug ) THEN
            PRINT '(A)','Bounds are acceptable for finding center latitude.'
         END IF
      ELSE
         PRINT '(A)','Bisection cannot converge as the solution is not contained within the bounds.'
         PRINT '(A,2F10.4,A,F10.4,A)','Guesses are:',fa,fb,'.  The value should be ',known_latitude,'.'
         error = .TRUE.
      END IF
      
      !  A couple of required initializations.
   
      converged = .FALSE.
      loop = 0
      xo = true_latitude
   
      !  Here is the loop to figure out the latitude.  We loop until
      !  the technique converges or until we have reached a maximum
      !  number of iterations.
   
      bisection : DO WHILE ( ( loop .LE. max_loops ) .AND. ( .NOT. converged ) .AND. &
                             ( .NOT. error ) ) 
         
         !  Increment the loop counter.
   
         loop = loop + 1
   
         !  Get the midpoint on the X-axis of the function.
   
         xn = ( xa + xb ) / 2.
         fn = xlat(r(known_x,point_x,known_y,point_y,grid_distance,c1,xn), &
              cone_factor,true_latitude)
   
         !  Which side is the solution contained within?
   
         IF      ( ( fn - known_latitude ) * ( fa - known_latitude ) .LE. 0 ) THEN
            xb = xn
            fb = fn
         ELSE IF ( ( fn - known_latitude ) * ( fb - known_latitude ) .LE. 0 ) THEN
            xa = xn
            fa = fn
         ELSE
            PRINT '(A)','We are not supposed to be able to get to here.'
            error = .TRUE.
            EXIT bisection
         END IF
   
         !  Have we converged yet?
   
         IF ( ABS(xn-xo) .LT. 0.001 ) THEN
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A,I4,A,F9.4,A)',&
               'Converged with ',fn,' in ',loop,' iterations, which is nearly = ',known_latitude,'.'
               PRINT '(A,F9.4,A)','The computed center latitude is ',xn,'.'
            END IF
            point_latitude = xn
            converged = .TRUE.
         ELSE
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A,F9.4)','No convergence with ',fn,', which is not quite = ',known_latitude
               PRINT '(A,F9.4,A)','The assumed center latitude is ',xn,'.'
            END IF
            xo=xn
            converged = .FALSE.
         END IF
   
      END DO bisection
      
   END SUBROUTINE center_latitude_lambert

   SUBROUTINE center_latitude_mercator ( known_latitude , true_latitude , center_longitude , &
                                         known_x , point_x , known_y , point_y , &
                                         grid_distance , point_latitude , converged )
   
      IMPLICIT NONE
   
      !  Input data.
   
      REAL , INTENT(IN)     :: known_latitude  , &  ! this value of latitude we know, usually lower left point
                               true_latitude   , &  ! the first true latitude
                               center_longitude, &  ! the longitude that is parallel to the y-axis
                               known_x         , &  ! x location of the known value of latitude, usually 1
                               point_x         , &  ! x location of point with unknown latitude, usually center
                               known_y         , &  ! y location of the known value of latitude, usually 1 
                               point_y         , &  ! y location of point with unknown latitude, usually center
                               grid_distance        ! grid distance at true latitude
   
      !  Output data.
   
      REAL , INTENT(OUT)    :: point_latitude       ! the value of the latitude at point_x, point_y
      LOGICAL , INTENT(OUT) :: converged

      !  Local data.

      REAL :: c2 , cell , yy
      REAL :: xa , xb , xn , xo
      REAL :: fa , fb , fn

      INTEGER :: loop
      INTEGER , PARAMETER :: max_loops = 1000

      LOGICAL :: error = .FALSE.

      !  Compute some constants from the input data.

      c2 = radius_of_earth * COS ( true_latitude * radians_per_degree )
      cell = COS ( known_latitude * radians_per_degree ) / &
             ( 1. + SIN ( known_latitude * radians_per_degree ) )
      yy = -c2 * LOG (cell)

      !  Compute the starting values of the bisection.

      xa = -80
      fa = ( EXP ( ( ( known_y -point_y ) * grid_distance - yy ) / c2 ) ) * &
           ( 1. + SIN ( xa * radians_per_degree ) ) - COS ( xa * radians_per_degree )

      xb =  80
      fb = ( EXP ( ( ( known_y -point_y ) * grid_distance - yy ) / c2 ) ) * &
           ( 1. + SIN ( xb * radians_per_degree ) ) - COS ( xb * radians_per_degree )

      !  First test to see that we have chosen reasonable bounds.

      IF ( fa * fb .LE. 0 ) THEN
         IF ( rec4%print_debug ) THEN
            PRINT '(A)','Bounds are acceptable for finding center latitude.'
         END IF
      ELSE
         PRINT '(A)','Bisection cannot converge as the solution is not contained within the bounds.'
         PRINT '(A,2F10.4,A)','Guesses are:',fa,fb,'.  The value should be 0.'
         error = .TRUE.
      END IF

      xo = xa
      loop = 0
      converged = .FALSE.
     
      bisection : DO WHILE ( ( loop .LE. max_loops ) .AND. ( .NOT. converged ) .AND. &
                             ( .NOT. error ) ) 
         
         !  Increment the loop counter.
   
         loop = loop + 1
   
         !  Get the midpoint on the X-axis of the function.
   
         xn = ( xa + xb ) / 2.
         fn = ( EXP ( ( ( known_y -point_y ) * grid_distance - yy ) / c2 ) ) * &
           ( 1. + SIN ( xn * radians_per_degree ) ) - COS ( xn * radians_per_degree )
   
         !  Which side is the solution contained within?
   
         IF      ( fn * fa .LE. 0 ) THEN
            xb = xn
            fb = fn
         ELSE IF ( fn * fb .LE. 0 ) THEN
            xa = xn
            fa = fn
         ELSE
            PRINT '(A)','We are not supposed to be able to get to here.'
            error = .TRUE.
            EXIT bisection
         END IF
   
         !  Have we converged yet?
   
         IF ( ABS(xn-xo) .LT. 0.00001 ) THEN
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A,I4,A)',&
               'Converged with ',fn,' in ',loop,' iterations, which is nearly = 0.'
               PRINT '(A,F9.4,A)','The computed center latitude is ',xn,'.'
            END IF
            point_latitude = xn
            converged = .TRUE.
         ELSE
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A)','No convergence with ',fn,', which is not quite = 0.'
               PRINT '(A,F9.4,A)','The assumed center latitude is ',xn,'.'
            END IF
            xo=xn
            converged = .FALSE.
         END IF
   
      END DO bisection
      
   END SUBROUTINE center_latitude_mercator

   SUBROUTINE center_latitude_polar ( known_latitude , true_latitude , center_longitude , &
                                      known_x , point_x , known_y , point_y , &
                                      grid_distance , point_latitude , converged )
   
      IMPLICIT NONE
   
      !  Input data.
   
      REAL , INTENT(IN)     :: known_latitude  , &  ! this value of latitude we know, usually lower left point
                               true_latitude   , &  ! the first true latitude
                               center_longitude, &  ! the longitude that is parallel to the y-axis
                               known_x         , &  ! x location of the known value of latitude, usually 1
                               point_x         , &  ! x location of point with unknown latitude, usually center
                               known_y         , &  ! y location of the known value of latitude, usually 1 
                               point_y         , &  ! y location of point with unknown latitude, usually center
                               grid_distance        ! grid distance at true latitude
   
      !  Output data.
   
      REAL , INTENT(OUT)    :: point_latitude       ! the value of the latitude at point_x, point_y
      LOGICAL , INTENT(OUT) :: converged
   
      !  Local data.
   
      REAL                  :: xlat , xx , yy , xxx , yyy , &
                               xa , xb, xn , fa , fb , fn , xo
   
      INTEGER               :: loop , max_loops = 100
   
      LOGICAL               :: error = .FALSE.
   
   
      !  This routine uses the bisection method to determine the latitude of 
      !  the center of the domain.  The input data is typical of what is available
      !  from the GRIB files for a polar stereographic projection.
   
      !  Statement functions for use in the iteration.
   
      xlat(xxx,yyy,true_latitude) = 90. - &
         2. * ATAN(1./radius_of_earth*SQRT(xxx**2+yyy**2)/ &
         (1.+COS((90.-true_latitude)*radians_per_degree)))/radians_per_degree
   
      yy(true_latitude,known_y,point_y,grid_distance,point_latitude)= &
         -1. * radius_of_earth * &
         SIN((90.-point_latitude)*radians_per_degree) * &
         (1.+COS((90.-true_latitude)*radians_per_degree))/ &
         (1.+COS((90.-point_latitude)*radians_per_degree)) + &
         (known_y - point_y) * grid_distance
   
      !  Initialize the output value to a recognizably undefined latitude.
   
      point_latitude = -999. 
      xx = ( known_x - point_x ) * grid_distance

   
      !  Compute the starting values of the bisection.  Again, I am assuming that this
      !  is not general enough for the southern hemisphere.
       
      xa = -89.
      fa = xlat(xx, &
                yy(true_latitude,known_y,point_y,grid_distance,xa),&
                true_latitude) 
      xb =  91.
      fb = xlat(xx, &
                yy(true_latitude,known_y,point_y,grid_distance,xb),&
                true_latitude) 
   
      !  First test to see that we have chosen reasonable bounds.
   
      IF ( ( fa - known_latitude ) * ( fb - known_latitude ) .LE. 0 ) THEN
         IF ( rec4%print_debug ) THEN
            PRINT '(A)','Bounds are acceptable for finding center latitude.'
         END IF
      ELSE
         PRINT '(A)','Bisection cannot converge as the solution is not contained within the bounds.'
         PRINT '(A,2F10.4,A,F10.4,A)','Guesses are:',fa,fb,'.  The value should be ',known_latitude,'.'
         error = .TRUE.
      END IF
      
      !  A couple of required initializations.
   
      converged = .FALSE.
      loop = 0
      xo = true_latitude
   
      !  Here is the loop to figure out the latitude.  We loop until
      !  the technique converges or until we have reached a maximum
      !  number of iterations.
   
      bisection : DO WHILE ( ( loop .LE. max_loops ) .AND. ( .NOT. converged ) .AND. &
                             ( .NOT. error ) ) 
         
         !  Increment the loop counter.
   
         loop = loop + 1
   
         !  Get the midpoint on the X-axis of the function.
   
         xn = ( xa + xb ) / 2.
         fn = xlat(xx, &
                yy(true_latitude,known_y,point_y,grid_distance,xn),&
                true_latitude) 
   
         !  Which side is the solution contained within?
   
         IF      ( ( fn - known_latitude ) * ( fa - known_latitude ) .LE. 0 ) THEN
            xb = xn
            fb = fn
         ELSE IF ( ( fn - known_latitude ) * ( fb - known_latitude ) .LE. 0 ) THEN
            xa = xn
            fa = fn
         ELSE
            PRINT '(A)','We are not supposed to be able to get to here.'
            error = .TRUE.
            EXIT bisection
         END IF
   
         !  Have we converged yet?
   
         IF ( ABS(xn-xo) .LT. 0.001 ) THEN
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A,I4,A,F9.4,A)',&
               'Converged with ',fn,' in ',loop,' iterations, which is nearly = ',known_latitude,'.'
               PRINT '(A,F9.4,A)','The computed center latitude is ',xn,'.'
            END IF
            point_latitude = xn
            point_latitude = MIN ( point_latitude , 90. ) 
            converged = .TRUE.
         ELSE
            IF ( rec4%print_debug ) THEN
               PRINT '(A,F9.4,A,F9.4)','No convergence with ',fn,', which is not quite = ',known_latitude
               PRINT '(A,F9.4,A)','The assumed center latitude is ',xn,'.'
            END IF
            xo=xn
            converged = .FALSE.
         END IF
   
      END DO bisection
      
   END SUBROUTINE center_latitude_polar
   
   SUBROUTINE lc_cone_factor (first_standard_parallel,second_standard_parallel,cone_factor)
   
      IMPLICIT NONE
   
      REAL    :: first_standard_parallel, & ! first standard latitude
                 second_standard_parallel   ! second standard latitude
      REAL    :: cone_factor                                
   
      !  Compute the cone factor for a Lambert conformal projection.  If there are
      !  two distinct standard parallels, we use the first (long) computation.  If 
      !  there is only a single standard parallel (typical of the projections in
      !  the GRIB document), we are talking SIN city.

      IF ( ABS ( first_standard_parallel - second_standard_parallel ) .GT. 0.1 ) THEN
         cone_factor = ALOG10(COS(first_standard_parallel  * radians_per_degree)) - &
                       ALOG10(COS(second_standard_parallel * radians_per_degree))
         cone_factor = cone_factor/(ALOG10(TAN((45.0 - ABS( first_standard_parallel)/2.0) * radians_per_degree)) - &
                                    ALOG10(TAN((45.0 - ABS(second_standard_parallel)/2.0) * radians_per_degree)))

      ELSE
         cone_factor = SIN ( ABS(first_standard_parallel) * radians_per_degree )

      END IF
   
   END SUBROUTINE lc_cone_factor

END MODULE first_guess_map
